#site-navigation{
	h1.site-logo {
		font-weight: bold;
	}
}

/* Add a line break between the arguments of each function.
   Taken from: https://github.com/sphinx-doc/sphinx/issues/1514#issuecomment-742703082
*/
.sig-param::before {
    content: "\a\20\20\20\20\20\20";
    white-space: pre;
}

dt em.sig-param:last-of-type::after {
    content: "\a";
    white-space: pre;
}

dl.class > dt:first-of-type {
    display: block !important;
}
